use std::{
    fs::File,
    io::BufReader,
    thread,
};
use std::collections::HashMap;
use std::io::{Read, Seek, SeekFrom};
use std::sync::mpsc;
use onebrc::{parse_line, parse_temp, read_bytes_until};

const DEFAULT_HASHMAP_LENGTH: usize = 10000;

fn main() {
    const FILE_PATH: &str = "../../../measurements.txt";
    let file = File::open(FILE_PATH).expect("File measurements.txt not found");
    thread::scope(|s| {
        let mut stations: HashMap<String, onebrc::StationMeasurements> = HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);
        let (tx, rx) = mpsc::channel();
        let cores = thread::available_parallelism().unwrap().into();
        let mut reader = BufReader::new(&file);
        let file_length = reader.seek(SeekFrom::End(0)).unwrap();
        let chunk_length = file_length as usize / cores;
        reader.seek(SeekFrom::Start(0)).unwrap();
        for i in 0..cores {
            let tx = tx.clone();
            s.spawn(move || {
                let file = File::open(FILE_PATH).expect("File measurements.txt not found");
                let mut reader = BufReader::new(&file);
                let mut currposition = (chunk_length * i) as u64;
                let end = currposition + chunk_length as u64;
                reader.seek(SeekFrom::Start(currposition)).unwrap();
                let mut bytes = reader.bytes();

                while let Some(byte) = bytes.next() {
                    match byte {
                        Ok(byte) => {
                            if byte == b'\n' {
                                break;
                            }
                        }
                        Err(_) => { panic!("could not go to next") }
                    }
                }

                let mut t_stations: HashMap<String, onebrc::StationMeasurements> =
                    HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);

                while let Some(line) = read_bytes_until(&mut bytes, b'\n') {
                    let (station, temp) = parse_line(&line);
                    let station = unsafe { String::from_utf8_unchecked(Vec::from(station)) };
                    let temp = parse_temp(temp);
                    let measurements_option = t_stations.get_mut(&station);
                    if let Some(measurements) = measurements_option {
                        measurements.update(temp);
                    } else {
                        let measurements = onebrc::StationMeasurements {
                            min: temp,
                            max: temp,
                            count: 1,
                            sum: temp,
                        };
                        t_stations.insert(station, measurements);
                    }
                    currposition += line.len() as u64;
                    if currposition > end {
                        break;
                    }
                }
                let _ = tx.send(t_stations);
            });
        }
        drop(tx);
        while let Ok(t_stations) = rx.recv() {
            for (station, measurements) in t_stations.iter() {
                let joined_measurements_options = stations.get_mut(station.as_str());
                if let Some(joined_measurements) = joined_measurements_options {
                    joined_measurements.merge(measurements);
                } else {
                    stations.insert(station.to_owned(), *measurements);
                }
            }
        }
        let mut stations: Vec<String> = stations.iter().map(|(station, measurements)| {
            let measurements = measurements.to_string();
            #[cfg(feature = "json")]
            {
                format!("{{\"{station}\":\"{measurements}\"}}")
            }
            #[cfg(not(feature = "json"))]
            {
                format!("{station}={measurements}")
            }
        }).collect();
        stations.sort();
        let stations = stations.join(",");
        #[cfg(feature = "json")]
        {
            println!("\n\n[{stations}]");
        }
        #[cfg(not(feature = "json"))]
        {
            println!("\n\n{{{stations}}}");
        }
        // println!("\n\nTime={} ms", now.elapsed().as_millis());
    });
}
