use std::{
    fs::File,
    io::{BufRead, BufReader},
    thread,
};
use std::collections::HashMap;
use std::sync::mpsc;
use std::time::Instant;

use onebrc::parse_temp;

const DEFAULT_HASHMAP_LENGTH: usize = 10000;

fn main() {
    //print!("\x1b[2J");
    print!("\x1b[s");
    thread::scope(|s| {
        let mut stations: HashMap<String, onebrc::StationMeasurements> = HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);
        let (tx, rx) = mpsc::channel();
        let now = Instant::now();
        let cores: usize = thread::available_parallelism().unwrap().into();
        let chunk_length = 1_000_000_000 / cores;
        for i in 0..cores {
            let print_line = i + 1;
            print!("\x1b[u\x1b[{print_line}B\x1b[0CThread #{i:0>2}: 0%");
            let file = File::open("../../../measurements.txt").expect("File measurements.txt not found");
            let reader = BufReader::new(file);
            let line_chunk = reader.lines().skip(chunk_length * i).take(chunk_length);
            let tx = tx.clone();
            s.spawn(move || {
                let mut t_stations: HashMap<String, onebrc::StationMeasurements> =
                    HashMap::with_capacity(DEFAULT_HASHMAP_LENGTH);

                let now_read_line = Instant::now();
                let mut line_num = 0;
                line_chunk.for_each(|line| {
                    if line_num == 0 {
                        print!("\x1b[u\x1b[{print_line}B\x1b[30CStart read line {}ms", now_read_line.elapsed().as_millis());
                    }
                    if line_num % 10000 == 0 {
                        //let formatted = format_nums(line_num);
                        //print!("\x1b[u\x1b[{print_line}B\x1b[0CThread #{i:0>2}: {formatted}");
                        let percent = (line_num as f64 / chunk_length as f64) * 100.0;
                        print!("\x1b[u\x1b[{print_line}B\x1b[0CThread #{i:0>2}: {percent:.2}%");
                    }
                    line_num += 1;
                    let line = line.expect("could not read line");
                    let (station, temp) = line.split_once(';').expect("Error while splitting");
                    let temp = parse_temp(temp.as_bytes());
                    let measurements_option = t_stations.get_mut(station);
                    if let Some(measurements) = measurements_option {
                        measurements.update(temp);
                    } else {
                        let measurements = onebrc::StationMeasurements {
                            min: temp,
                            max: temp,
                            count: 1,
                            sum: temp,
                        };
                        t_stations.insert(station.to_owned(), measurements);
                    }
                });
                print!("\x1b[u\x1b[{print_line}B\x1b[60CTime reading lines in thread {i}={} ms", now_read_line.elapsed().as_millis());
                let _ = tx.send(t_stations);
            });
        }
        print!("\x1b[{cores}B");
        drop(tx);
        while let Ok(t_stations) = rx.recv() {
            for (station, measurements) in t_stations.iter() {
                let joined_measurements_options = stations.get_mut(station.as_str());
                if let Some(joined_measurements) = joined_measurements_options {
                    joined_measurements.merge(measurements);
                } else {
                    stations.insert(station.to_owned(), *measurements);
                }
            }
        }
        let mut stations: Vec<String> = stations.iter().map(|(station, measurements)| {
            let measurements = measurements.to_string();
            format!("{station}={measurements}")
        }).collect();
        stations.sort();
        let _stations = stations.join(",");
        // println!("{{{stations}}}");
        println!("\n\nTime={} ms", now.elapsed().as_millis());
    });
}
